/* --------------------------------------------------------------  */
/* (C)Copyright 2007                                               */
/* International Business Machines Corporation                     */
/* All Rights Reserved.                                            */
/*                                                                 */
/* Redistribution and use in source and binary forms, with or      */
/* without modification, are permitted provided that the           */
/* following conditions are met:                                   */
/*                                                                 */
/* - Redistributions of source code must retain the above copyright*/
/*   notice, this list of conditions and the following disclaimer. */
/*                                                                 */
/* - Redistributions in binary form must reproduce the above       */
/*   copyright notice, this list of conditions and the following   */
/*   disclaimer in the documentation and/or other materials        */
/*   provided with the distribution.                               */
/*                                                                 */
/* - Neither the name of IBM Corporation nor the names of its      */
/*   contributors may be used to endorse or promote products       */
/*   derived from this software without specific prior written     */
/*   permission.                                                   */
/*                                                                 */
/* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND          */
/* CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,     */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF        */
/* MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE        */
/* DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR            */
/* CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,    */
/* SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT    */
/* NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;    */
/* LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)        */
/* HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN       */
/* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR    */
/* OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,  */
/* EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.              */
/* --------------------------------------------------------------  */
/* PROLOG END TAG zYx                                              */
#ifndef _ASINHF4_H_
#define _ASINHF4_H_	1

#include <altivec.h>
#include <vec_types.h>

#include "logf4.h"
#include "sqrtf4.h"

/*
 * Maclaurin Series Coefficients 
 * for x near 0.
 */
#define ASINH_MAC01     1.0000000000000000000000000000000000000000000000000000000000000000000000E0f
#define ASINH_MAC03    -1.6666666666666666666666666666666666666666666666666666666666666666666667E-1f
#define ASINH_MAC05     7.5000000000000000000000000000000000000000000000000000000000000000000000E-2f
#define ASINH_MAC07    -4.4642857142857142857142857142857142857142857142857142857142857142857143E-2f
#define ASINH_MAC09     3.0381944444444444444444444444444444444444444444444444444444444444444444E-2f
#define ASINH_MAC11    -2.2372159090909090909090909090909090909090909090909090909090909090909091E-2f
#define ASINH_MAC13     1.7352764423076923076923076923076923076923076923076923076923076923076923E-2f
#define ASINH_MAC15    -1.3964843750000000000000000000000000000000000000000000000000000000000000E-2f
#define ASINH_MAC17     1.1551800896139705882352941176470588235294117647058823529411764705882353E-2f
#define ASINH_MAC19    -9.7616095291940789473684210526315789473684210526315789473684210526315789E-3f
#define ASINH_MAC21     8.3903358096168154761904761904761904761904761904761904761904761904761905E-3f
#define ASINH_MAC23    -7.3125258735988451086956521739130434782608695652173913043478260869565217E-3f
#define ASINH_MAC25     6.4472103118896484375000000000000000000000000000000000000000000000000000E-3f
#define ASINH_MAC27    -5.7400376708419234664351851851851851851851851851851851851851851851851852E-3f
#define ASINH_MAC29     5.1533096823199041958512931034482758620689655172413793103448275862068966E-3f


/*
 * FUNCTION
 *  vector float _asinhf4(vector float x)
 *
 * DESCRIPTION
 *  The asinhf4 function returns a vector containing the hyperbolic
 *  arcsines of the corresponding elements of the input vector.
 *
 *  We are using the formula:
 *    asinh = ln(|x| + sqrt(x^2 + 1))
 *  and the anti-symmetry of asinh.
 *
 *  For x near zero, we use the Taylor series:
 *
 *                infinity
 *                 ------
 *                  -   '  P  (0)
 *                   -      k-1    k
 *    asinh x =       -    -----  x
 *                   -       k
 *                  -   ,
 *                 ------
 *                 k = 1
 *
 *  Special Cases:
 *   - asinh(+0)        returns +0
 *   - asinh(-0)        returns -0
 *   - Normally,  asinh(+/- infinity) returns +/- infinity,
 *	   but on the SPU, single-precision infinity is not supported,
 *	   so it is treated as a normal number here.
 *
 */

static __inline vector float _asinhf4(vector float x)
{
    vec_float4 sign_mask = (vector float){-0.0f, -0.0f, -0.0f, -0.0f};
    vec_float4 zerof     = (vector float){0.0f, 0.0f, 0.0f, 0.0f};
    vec_float4 onef      = (vector float){1.0f, 1.0f, 1.0f, 1.0f};
    vec_float4 result, fresult, mresult;;
    vec_float4 xabs, xsqu;
    /* Where we switch from maclaurin to formula */
    vec_float4  switch_approx = (vector float){0.685f, 0.685f, 0.685f, 0.685f};
    vector bool int use_form;

   
    xabs = vec_andc(x, sign_mask);
    xsqu = vec_madd(x, x, zerof);

    /*
     * Formula:
     *   asinh = ln(|x| + sqrt(x^2 + 1))
     */
    fresult = _sqrtf4(vec_madd(x, x, onef));
    fresult = vec_add(xabs, fresult);
    fresult = _logf4(fresult);


    /*
     * Maclaurin Series
     */
    mresult = vec_madd(xsqu, (vector float){ASINH_MAC29, ASINH_MAC29, ASINH_MAC29, ASINH_MAC29}, 
                             (vector float){ASINH_MAC27, ASINH_MAC27, ASINH_MAC27, ASINH_MAC27});
    mresult = vec_madd(xsqu, mresult, (vector float){ASINH_MAC25, ASINH_MAC25, ASINH_MAC25, ASINH_MAC25});
    mresult = vec_madd(xsqu, mresult, (vector float){ASINH_MAC23, ASINH_MAC23, ASINH_MAC23, ASINH_MAC23});
    mresult = vec_madd(xsqu, mresult, (vector float){ASINH_MAC21, ASINH_MAC21, ASINH_MAC21, ASINH_MAC21});
    mresult = vec_madd(xsqu, mresult, (vector float){ASINH_MAC19, ASINH_MAC19, ASINH_MAC19, ASINH_MAC19});
    mresult = vec_madd(xsqu, mresult, (vector float){ASINH_MAC17, ASINH_MAC17, ASINH_MAC17, ASINH_MAC17});
    mresult = vec_madd(xsqu, mresult, (vector float){ASINH_MAC15, ASINH_MAC15, ASINH_MAC15, ASINH_MAC15});
    mresult = vec_madd(xsqu, mresult, (vector float){ASINH_MAC13, ASINH_MAC13, ASINH_MAC13, ASINH_MAC13});
    mresult = vec_madd(xsqu, mresult, (vector float){ASINH_MAC11, ASINH_MAC11, ASINH_MAC11, ASINH_MAC11});
    mresult = vec_madd(xsqu, mresult, (vector float){ASINH_MAC09, ASINH_MAC09, ASINH_MAC09, ASINH_MAC09});
    mresult = vec_madd(xsqu, mresult, (vector float){ASINH_MAC07, ASINH_MAC07, ASINH_MAC07, ASINH_MAC07});
    mresult = vec_madd(xsqu, mresult, (vector float){ASINH_MAC05, ASINH_MAC05, ASINH_MAC05, ASINH_MAC05});
    mresult = vec_madd(xsqu, mresult, (vector float){ASINH_MAC03, ASINH_MAC03, ASINH_MAC03, ASINH_MAC03});
    mresult = vec_madd(xsqu, mresult, (vector float){ASINH_MAC01, ASINH_MAC01, ASINH_MAC01, ASINH_MAC01});
    mresult = vec_madd(xabs, mresult, zerof);


    /*
     * Choose between series and formula
     */
    use_form = vec_cmpgt(xabs, switch_approx);
    result = vec_sel(mresult, fresult, use_form);


    /* Preserve sign - asinh is anti-symmetric */
    result = vec_sel(result, x, (vec_uint4)sign_mask);

    return result;
}

#endif /* _ASINHF4_H_ */
