/* --------------------------------------------------------------  */
/* (C)Copyright 2007                                               */
/* International Business Machines Corporation                     */
/* All Rights Reserved.                                            */
/*                                                                 */
/* Redistribution and use in source and binary forms, with or      */
/* without modification, are permitted provided that the           */
/* following conditions are met:                                   */
/*                                                                 */
/* - Redistributions of source code must retain the above copyright*/
/*   notice, this list of conditions and the following disclaimer. */
/*                                                                 */
/* - Redistributions in binary form must reproduce the above       */
/*   copyright notice, this list of conditions and the following   */
/*   disclaimer in the documentation and/or other materials        */
/*   provided with the distribution.                               */
/*                                                                 */
/* - Neither the name of IBM Corporation nor the names of its      */
/*   contributors may be used to endorse or promote products       */
/*   derived from this software without specific prior written     */
/*   permission.                                                   */
/*                                                                 */
/* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND          */
/* CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,     */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF        */
/* MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE        */
/* DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR            */
/* CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,    */
/* SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT    */
/* NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;    */
/* LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)        */
/* HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN       */
/* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR    */
/* OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,  */
/* EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.              */
/* --------------------------------------------------------------  */
/* PROLOG END TAG zYx                                              */
#ifndef _ERFCF4_H_
#define _ERFCF4_H_	1

#include <altivec.h>
#include <vec_types.h>

#include "expf4.h"
#include "recipf4.h"
#include "divf4.h"
#include "erff4.h"

/*
 * FUNCTION
 *  vector float _erfcf4(vector float x)
 *
 * DESCRIPTION
 *  The erfcf4 function computes the complement error function of each element of x.
 *
 *  C99 Special Cases:
 *  - erfc(+0) returns +1
 *  - erfc(-0) returns +1
 *  - erfc(+infinite) returns +0
 *  - erfc(-infinite) returns +2
 *
 */

static __inline vector float _erfcf4(vector float x)
{
  vec_float4 onehalff  = (vector float){0.5f, 0.5f, 0.5f, 0.5f};
  vec_float4 zerof     = (vector float){0.0f, 0.0f, 0.0f, 0.0f};
  vec_float4 onef      = (vector float){1.0f, 1.0f, 1.0f, 1.0f};
  vec_float4 twof      = (vector float){2.0f, 2.0f, 2.0f, 2.0f};
  vec_float4 sign_mask = (vector float){-0.0f, -0.0f, -0.0f, -0.0f};

  vec_float4 infp = (vector float) ((vector unsigned int){0x7F800000, 0x7F80000, 0x7F80000, 0x7F80000});
  vec_float4 infn = (vector float) ((vector unsigned int){0xFF800000, 0xFF80000, 0xFF80000, 0xFF80000});

  /* This is where we switch from near zero approx. */
  vec_float4 approx_point = (vector float){0.89f, 0.89f, 0.89f, 0.89f};

  vector bool  isneg;
  vec_float4 xabs, xsqu, xsign;
  vec_float4 tresult, presult, result;

  xsign = vec_and(x, sign_mask);

  /* Force Denorms to 0 */
  x = vec_add(x, zerof);

  xabs = vec_andc(x, sign_mask);
  xsqu = vec_madd(x, x, zerof);

  /*
   * Use Taylor Series for x near 0
   * Preserve sign of x in result, since erf(-x) = -erf(x)
   * This approximation is for erf, so adjust for erfc.
   */
  TAYLOR_ERFF4(xabs, xsqu, tresult);
  tresult = vec_or(tresult, xsign);
  tresult = vec_sub(onef, tresult);

  /*
   * Now, use the Continued Fractions approximation away
   * from 0. If x < 0, use erfc(-x) = 2 - erfc(x)
   */
  CONTFRAC_ERFCF4(xabs, xsqu, presult);

  isneg = vec_cmplt(x, zerof);
  presult = vec_sel(presult, vec_sub(twof, presult), isneg);

  /*
   * Select the appropriate approximation.
   */
  result = vec_sel(tresult, presult, vec_cmpgt(xabs, approx_point));

  /*
   * Special cases/errors.
   */

  /* x = +/- infinite */
  result = vec_sel(result, zerof, vec_cmpeq(xabs, infp));
  result = vec_sel(result, twof,  vec_cmpeq(xabs, infn));

  return result;
}

#endif /* _ERFCF4_H_ */
