/* --------------------------------------------------------------  */
/* (C)Copyright 2007                                               */
/* International Business Machines Corporation                     */
/* All Rights Reserved.                                            */
/*                                                                 */
/* Redistribution and use in source and binary forms, with or      */
/* without modification, are permitted provided that the           */
/* following conditions are met:                                   */
/*                                                                 */
/* - Redistributions of source code must retain the above copyright*/
/*   notice, this list of conditions and the following disclaimer. */
/*                                                                 */
/* - Redistributions in binary form must reproduce the above       */
/*   copyright notice, this list of conditions and the following   */
/*   disclaimer in the documentation and/or other materials        */
/*   provided with the distribution.                               */
/*                                                                 */
/* - Neither the name of IBM Corporation nor the names of its      */
/*   contributors may be used to endorse or promote products       */
/*   derived from this software without specific prior written     */
/*   permission.                                                   */
/*                                                                 */
/* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND          */
/* CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,     */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF        */
/* MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE        */
/* DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR            */
/* CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,    */
/* SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT    */
/* NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;    */
/* LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)        */
/* HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN       */
/* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR    */
/* OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,  */
/* EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.              */
/* --------------------------------------------------------------  */
/* PROLOG END TAG zYx                                              */

#ifndef _LGAMMAF4_H_
#define _LGAMMAF4_H_	1

#include <altivec.h>
#include <vec_types.h>

#include "simdmath.h"
#include "recipf4.h"
#include "logf4.h"
#include "sinf4.h"
#include "truncf4.h"



#define HALFLOG2PI          9.189385332046727417803297E-1f
#define EULER_MASCHERONI    0.5772156649015328606065f

/*
 * Zeta constants for Maclaurin approx. near zero
 */
#define ZETA_02_DIV_02       8.2246703342411321823620758332301E-1f
#define ZETA_03_DIV_03      -4.0068563438653142846657938717048E-1f
#define ZETA_04_DIV_04       2.7058080842778454787900092413529E-1f
#define ZETA_05_DIV_05      -2.0738555102867398526627309729141E-1f
#define ZETA_06_DIV_06       1.6955717699740818995241965496515E-1f

/*
 * Coefficients for Stirling's Series for Lgamma()
 */
#define STIRLING_01    8.3333333333333333333333333333333333333333333333333333333333333333333333E-2f
#define STIRLING_02   -2.7777777777777777777777777777777777777777777777777777777777777777777778E-3f
#define STIRLING_03    7.9365079365079365079365079365079365079365079365079365079365079365079365E-4f
#define STIRLING_04   -5.9523809523809523809523809523809523809523809523809523809523809523809524E-4f
#define STIRLING_05    8.4175084175084175084175084175084175084175084175084175084175084175084175E-4f
#define STIRLING_06   -1.9175269175269175269175269175269175269175269175269175269175269175269175E-3f
#define STIRLING_07    6.4102564102564102564102564102564102564102564102564102564102564102564103E-3f
#define STIRLING_08   -2.9550653594771241830065359477124183006535947712418300653594771241830065E-2f
#define STIRLING_09    1.7964437236883057316493849001588939669435025472177174963552672531000704E-1f
#define STIRLING_10   -1.3924322169059011164274322169059011164274322169059011164274322169059011E0f


/*
 * FUNCTION
 *	vector float _lgammaf4(vector float x) - Natural Log of Gamma Function
 *
 * DESCRIPTION
 *	_lgammaf4 calculates the natural logarithm of the absolute value of the gamma
 *	function for the corresponding elements of the input vector.
 *
 * C99 Special Cases:
 *	lgamma(0) returns +infinite 
 *	lgamma(1) returns +0
 *	lgamma(2) returns +0
 *	lgamma(negative integer) returns +infinite 
 *	lgamma(+infinite) returns +infinite
 *	lgamma(-infinite) returns +infinite
 *
 * Other Cases:
 *  lgamma(Nan) returns Nan
 *  lgamma(Denorm) treated as lgamma(0) and returns +infinite
 *
 */

static __inline vector float _lgammaf4(vector float x) 
{
  vec_float4 z2 = (vector float){ZETA_02_DIV_02, ZETA_02_DIV_02, ZETA_02_DIV_02, ZETA_02_DIV_02};
  vec_float4 z3 = (vector float){ZETA_03_DIV_03, ZETA_03_DIV_03, ZETA_03_DIV_03, ZETA_03_DIV_03};
  vec_float4 z4 = (vector float){ZETA_04_DIV_04, ZETA_04_DIV_04, ZETA_04_DIV_04, ZETA_04_DIV_04};
  vec_float4 z5 = (vector float){ZETA_05_DIV_05, ZETA_05_DIV_05, ZETA_05_DIV_05, ZETA_05_DIV_05};
  vec_float4 z6 = (vector float){ZETA_06_DIV_06, ZETA_06_DIV_06, ZETA_06_DIV_06, ZETA_06_DIV_06};

  vec_float4 st1  = (vector float){STIRLING_01, STIRLING_01, STIRLING_01, STIRLING_01};
  vec_float4 st2  = (vector float){STIRLING_02, STIRLING_02, STIRLING_02, STIRLING_02};
  vec_float4 st3  = (vector float){STIRLING_03, STIRLING_03, STIRLING_03, STIRLING_03};
  vec_float4 st4  = (vector float){STIRLING_04, STIRLING_04, STIRLING_04, STIRLING_04};
  vec_float4 st5  = (vector float){STIRLING_05, STIRLING_05, STIRLING_05, STIRLING_05};
  vec_float4 st6  = (vector float){STIRLING_06, STIRLING_06, STIRLING_06, STIRLING_06};
  vec_float4 st7  = (vector float){STIRLING_07, STIRLING_07, STIRLING_07, STIRLING_07};
  vec_float4 st8  = (vector float){STIRLING_08, STIRLING_08, STIRLING_08, STIRLING_08};
  vec_float4 st9  = (vector float){STIRLING_09, STIRLING_09, STIRLING_09, STIRLING_09};
  vec_float4 st10 = (vector float){STIRLING_10, STIRLING_10, STIRLING_10, STIRLING_10};

  vec_float4 pi         = (vec_float4){(float)SM_PI, (float)SM_PI, (float)SM_PI, (float)SM_PI};
  vec_float4 halflogpi  = (vec_float4){HALFLOG2PI, HALFLOG2PI, HALFLOG2PI, HALFLOG2PI};
  vec_float4 euler      = (vec_float4){EULER_MASCHERONI, EULER_MASCHERONI, EULER_MASCHERONI, EULER_MASCHERONI};

  vec_float4 halff = (vec_float4) {0.5f, 0.5f, 0.5f, 0.5f};
  vec_float4 sixf = (vec_float4) {6.0f, 6.0f, 6.0f, 6.0f};

  vec_float4 inff       = (vec_float4)((vector unsigned int){0x7F800000, 0x7F800000, 0x7F800000, 0x7F800000});
  vec_float4 zerof      = (vec_float4){0.0f, 0.0f, 0.0f, 0.0f};
  vec_float4 sign_maskf = (vec_float4)((vector unsigned int){0x80000000, 0x80000000, 0x80000000, 0x80000000});

  vector bool int gt0;

  /* This is where we switch from near zero approx. */
  vec_float4 mac_switch   = (vec_float4){0.16f, 0.16f, 0.16f, 0.16f};
  vec_float4 shift_switch = (vec_float4){6.0f,  6.0f,  6.0f,  6.0f};

  vec_float4 inv_x, inv_xsqu;                  
  vec_float4 xtrunc, xstirling;
  vec_float4 sum, xabs;
  vector bool int  isshifted;
  vec_float4 result, stresult, shresult, mresult, nresult;


  xabs = vec_andc(x, sign_maskf);

  gt0    = vec_cmpgt(x, zerof);
  xtrunc = _truncf4(x);

  /*
   * For 0 < x <= 0.16.
   * Approximation Near Zero
   *
   * Use Maclaurin Expansion of lgamma()
   *
   * lgamma(z) = -ln(z) - z * EulerMascheroni + Sum[(-1)^n * z^n * Zeta(n)/n]
   */
  mresult = vec_madd(xabs, z6, z5);
  mresult = vec_madd(xabs, mresult, z4);
  mresult = vec_madd(xabs, mresult, z3);
  mresult = vec_madd(xabs, mresult, z2);
  mresult = vec_madd(xabs, vec_madd(xabs, mresult, zerof), zerof);

  mresult = vec_sub(mresult, vec_add(_logf4(xabs), vec_madd(xabs, euler, zerof)));


  /*
   * For 0.16 < x <= 6.0, we are going to push value
   * out to an area where Stirling's approximation is
   * accurate. Let's use a constant of 6.
   *
   * Use the recurrence relation:
   *    lgamma(x + 1) = ln(x) + lgamma(x)
   * 
   * Note that we shift x here, before Stirling's calculation,
   * then after Stirling's, we adjust the result.
   *
   */
  isshifted = vec_cmpgt(shift_switch, x);
  xstirling = vec_sel(xabs, vec_add(xabs, sixf), isshifted);
  inv_x    = _recipf4(xstirling);            
  inv_xsqu = vec_madd(inv_x, inv_x, zerof);            

  /*
   * For 6.0 < x < infinite
   *
   * Use Stirling's Series.
   *
   *              1                    1                1      1        1
   * lgamma(x) = --- ln (2*pi) + (z - ---) ln(x) - x + --- - ----- + ------ ...
   *              2                    2               12x   360x^3  1260x^5
   *
   *
   */
  sum = vec_madd(inv_xsqu, st10, st9);
  sum = vec_madd(sum, inv_xsqu, st8);
  sum = vec_madd(sum, inv_xsqu, st7);
  sum = vec_madd(sum, inv_xsqu, st6);
  sum = vec_madd(sum, inv_xsqu, st5);
  sum = vec_madd(sum, inv_xsqu, st4);
  sum = vec_madd(sum, inv_xsqu, st3);
  sum = vec_madd(sum, inv_xsqu, st2);
  sum = vec_madd(sum, inv_xsqu, st1);
  sum = vec_madd(sum, inv_x, zerof);

  stresult = vec_madd(vec_sub(xstirling, halff), _logf4(xstirling), halflogpi);
  stresult = vec_sub(stresult, xstirling);
  stresult = vec_add(stresult, sum);

  /*
   * Adjust result if we shifted x into Stirling range.
   *
   * lgamma(x) = lgamma(x + n) - ln(x(x+1)(x+2)...(x+n-1)
   *
   */
  shresult = vec_madd(xabs,     vec_add(xabs, (vec_float4) {1.0f, 1.0f, 1.0f, 1.0f}), zerof);
  shresult = vec_madd(shresult, vec_add(xabs, (vec_float4) {2.0f, 2.0f, 2.0f, 2.0f}), zerof);
  shresult = vec_madd(shresult, vec_add(xabs, (vec_float4) {3.0f, 3.0f, 3.0f, 3.0f}), zerof);
  shresult = vec_madd(shresult, vec_add(xabs, (vec_float4) {4.0f, 4.0f, 4.0f, 4.0f}), zerof);
  shresult = vec_madd(shresult, vec_add(xabs, (vec_float4) {5.0f, 5.0f, 5.0f, 5.0f}), zerof);
  shresult = _logf4(shresult);
  shresult = vec_sub(stresult, shresult);
  stresult = vec_sel(stresult, shresult, isshifted);


  /*
   * Select either Maclaurin or Stirling result before Negative X calc.
   */
//  vec_uint4 useStirlings = spu_cmpgt(xabs, mac_switch);
  result = vec_sel(mresult, stresult, vec_cmpgt(xabs, mac_switch));

  /*
   * Approximation for Negative X
   *
   * Use reflection relation:
   *
   * gamma(x) * gamma(-x) = -pi/(x sin(pi x))
   *
   * lgamma(x) = log(pi/(-x sin(pi x))) - lgamma(-x)
   *           
   */
  nresult = vec_madd(x, _sinf4(vec_madd(x, pi, zerof)), zerof);
  nresult = vec_andc(nresult, sign_maskf);
  nresult = vec_sub(_logf4(pi), vec_add(result, _logf4(nresult)));

  /*
   * Select between the negative or positive x approximations.
   */
  result = vec_sel(nresult, result, gt0);

  /*
   * Finally, special cases/errors.
   */

  /*
   * x = non-positive integer, return infinity.
   */
  result = vec_sel(result, inff, vec_andc(vec_cmpeq(x, xtrunc), gt0));

  /* x = +/- infinite or nan, return |x| */
  result   = vec_sel(result, xabs, vec_cmpgt((vec_uint4)xabs, (vec_uint4) {0x7FEFFFFF, 0x7FEFFFFF, 0x7FEFFFFF, 0x7FEFFFFF}));

  return result;
}

#endif /* _LGAMMAF4_H_ */



